import random
from typing import Tuple, Optional, Dict, Any, cast, List

import numpy as np

from .Unit import Unit
from .ILayerListener import ILayerListener
from ..IDataTransfer import IDataTransfer
from ..Listenable import Listenable
from ..constants import Direction, CellValue


class Layer(Listenable[ILayerListener], IDataTransfer):
    def __init__(self, size: Tuple[int, int], defaultValue: int = 0):
        super().__init__()
        self.__size = size
        self.__defaultValue = int(defaultValue)
        self.__cells = np.full([size[0] + 2, size[1] + 2], defaultValue, dtype=np.int32)
        self.__units: Dict[Tuple[int, int], Unit] = {}

    def __eq__(self, other) -> bool:
        if not isinstance(other, Layer):
            return False
        layer = cast(Layer, other)
        if self.__size != layer.__size:
            return False
        if self.__defaultValue != layer.__defaultValue:
            return False
        if not np.array_equal(self.__cells, layer.__cells):
            return False
        if len(self.__units) != len(layer.__units):
            return False
        for coords, unit in self.__units.items():
            if coords not in layer.__units:
                return False
            if unit != layer.__units[coords]:
                return False
        return True

    @property
    def size(self) -> Tuple[int, int]:
        return self.__size

    @property
    def width(self) -> int:
        return self.__size[0]

    @property
    def height(self) -> int:
        return self.__size[1]

    @property
    def cells(self) -> np.ndarray:
        return self.__cells[1:-1, 1:-1]

    def __getitem__(self, coords: Tuple[int, int]) -> int:
        return self.getValue(coords)

    def getValue(self, coords: Tuple[int, int], direction: Optional[Direction] = None) -> int:
        x, y = coords[0], coords[1]
        w, h = self.__size
        assert 0 <= x < w, f"Invalid x={x}"
        assert 0 <= y < h, f"Invalid y={y}"
        if direction:
            if direction == Direction.LEFT:
                return int(self.__cells[x, y + 1])
            if direction == Direction.RIGHT:
                return int(self.__cells[x + 2, y + 1])
            if direction == Direction.TOP:
                return int(self.__cells[x + 1, y])
            if direction == Direction.BOTTOM:
                return int(self.__cells[x + 1, y + 2])
        return int(self.__cells[x + 1, y + 1])

    def __setitem__(self, coords: Tuple[int, int], value: int):
        self.setValue(coords, value)

    def setValue(self, coords: Tuple[int, int], value: int):
        x, y = coords[0], coords[1]
        assert 0 <= x < self.__size[0], f"Invalid x={x}"
        assert 0 <= y < self.__size[1], f"Invalid y={y}"
        assert coords not in self.__units, f"There is a unit at {x},{y}"
        self.__cells[x + 1, y + 1] = value

    def fill(self, value: int):
        self.__cells[1:-1, 1:-1] = value

    def getNeighbors4(self, cell: Tuple[int, int]) -> Tuple[int, int, int, int]:
        x, y = cell
        return int(self.__cells[x, y + 1]), \
               int(self.__cells[x + 1, y]), \
               int(self.__cells[x + 1, y + 2]), \
               int(self.__cells[x + 2, y + 1])

    def getArea(self, area: Tuple[int, int, int, int]) -> np.ndarray:
        minX, minY, maxX, maxY = area
        return self.__cells[minX + 1:maxX + 1, minY + 1:maxY + 1]

    def getAllNeighbors4(self) -> np.ndarray:
        w, h = self.__size
        top = self.__cells[1:w + 1, 0:h]
        left = self.__cells[0:w, 1:h + 1]
        right = self.__cells[2:w + 2, 1:h + 1]
        bottom = self.__cells[1:w + 1, 2:h + 2]
        return np.stack((left, top, bottom, right), axis=2)

    def getAreaNeighbors4(self, cellsBox: Tuple[int, int, int, int]) -> np.ndarray:
        minX, minY, maxX, maxY = cellsBox
        top = self.__cells[minX + 1:maxX + 1, minY:maxY]
        left = self.__cells[minX:maxX, minY + 1:maxY + 1]
        right = self.__cells[minX + 2:maxX + 2, minY + 1:maxY + 1]
        bottom = self.__cells[minX + 1:maxX + 1, minY + 2:maxY + 2]
        return np.stack((left, top, bottom, right), axis=2)

    def getNeighbors8(self, cell: Tuple[int, int]) -> Tuple[int, int, int, int, int, int, int, int]:
        x, y = cell
        n = np.delete(self.__cells[x:x + 3, y:y + 3].flatten(), 4)
        return int(n[0]), int(n[1]), int(n[2]), int(n[3]), \
               int(n[4]), int(n[5]), int(n[6]), int(n[7])

    def getAreaNeighbors8(self, cellsBox: Tuple[int, int, int, int]) -> np.ndarray:
        minX, minY, maxX, maxY = cellsBox
        topLeft = self.__cells[minX:maxX, minY:maxY]
        top = self.__cells[minX + 1:maxX + 1, minY:maxY]
        topRight = self.__cells[minX + 2:maxX + 2, minY:maxY]
        left = self.__cells[minX:maxX, minY + 1:maxY + 1]
        right = self.__cells[minX + 2:maxX + 2, minY + 1:maxY + 1]
        bottomLeft = self.__cells[minX:maxX, minY + 2:maxY + 2]
        bottom = self.__cells[minX + 1:maxX + 1, minY + 2:maxY + 2]
        bottomRight = self.__cells[minX + 2:maxX + 2, minY + 2:maxY + 2]
        return np.stack((topLeft, left, bottomLeft, top,
                         bottom, topRight, right, bottomRight), axis=2)

    def getAllNeighbors8(self) -> np.ndarray:
        w, h = self.__size
        topLeft = self.__cells[0:w, 0:h]
        top = self.__cells[1:w + 1, 0:h]
        topRight = self.__cells[2:w + 2, 0:h]
        left = self.__cells[0:w, 1:h + 1]
        right = self.__cells[2:w + 2, 1:h + 1]
        bottomLeft = self.__cells[0:w, 2:h + 2]
        bottom = self.__cells[1:w + 1, 2:h + 2]
        bottomRight = self.__cells[2:w + 2, 2:h + 2]
        return np.stack((topLeft, left, bottomLeft, top,
                         bottom, topRight, right, bottomRight), axis=2)

    # Units

    @property
    def units(self) -> Dict[Tuple[int, int], Unit]:
        return self.__units

    def getPlayerUnits(self, playerId: int) -> Dict[Tuple[int, int], Unit]:
        return {
            coords: unit
            for coords, unit in self.__units.items()
            if unit.playerId == playerId
        }

    def getUnit(self, coords: Tuple[int, int]) -> Optional[Unit]:
        if coords not in self.__units:
            return None
        return self.__units[coords]

    def setUnit(self, coords: Tuple[int, int], value: int, unit: Optional[Unit] = None):
        x, y = coords[0], coords[1]
        assert 0 <= x < self.__size[0], f"Invalid x={x}"
        assert 0 <= y < self.__size[1], f"Invalid y={y}"
        if value == 0 or unit is None:
            self.__cells[x + 1, y + 1] = 0
            if coords in self.__units:
                del self.__units[coords]
        else:
            self.__cells[x + 1, y + 1] = value
            self.__units[coords] = unit

    # Layer listener

    def notifyContentChanged(self):
        for listener in self.listeners:
            listener.contentChanged(self)

    def notifyCellChanged(self, cell: Tuple[int, int]):
        for listener in self.listeners:
            listener.cellChanged(self, cell)

    # Data transfer

    def gatherData(self) -> Dict[str, Any]:
        randomShift = random.randint(0, 100)  # This is only for checking
        cellsValues = {
            str(self.__defaultValue + randomShift): CellValue(self.__defaultValue).toName()
        }
        for value in np.unique(self.__cells):
            cellsValues[str(value + randomShift)] = CellValue(value).toName()

        nonzeroX, nonzeroY = np.nonzero(self.__cells != self.__defaultValue)
        if len(nonzeroX) >= 0.1 * self.__cells.size:
            cellsData = {
                "format": "dense",
                "data": (self.__cells + randomShift).tolist(),
                "values": cellsValues
            }
        else:
            data = []
            for x, y in zip(nonzeroX, nonzeroY):
                value = self.__cells[x, y] + randomShift
                data.append([int(x), int(y), int(value)])
            cellsData = {
                "format": "sparse",
                "data": data,
                "values": cellsValues
            }

        unitsObjects = {}
        for unit in self.__units.values():
            unitsObjects[str(id(unit))] = unit.gatherData()
        unitsCell = []
        for coords, unit in self.__units.items():
            unitsCell.append([
                coords[0], coords[1], id(unit)
            ])
        unitsData = {
            "data": unitsObjects,
            "cells": unitsCell
        }

        return {
            "width": self.width,
            "height": self.height,
            "defaultValue": int(self.__defaultValue),
            "cells": cellsData,
            "units": unitsData
        }

    def takeData(self, data: Dict[str, Any]):
        cellsData = data["cells"]
        self.__size = (int(data["width"]), int(data["height"]))
        self.__defaultValue = int(data["defaultValue"])
        if cellsData["format"] == "dense":
            denseData = cellsData["data"]
            assert len(denseData) == self.width + 2
            assert len(denseData[0]) == self.height + 2
            self.__cells = np.array(denseData, dtype=np.int32)
        elif cellsData["format"] == "sparse":
            self.__cells = np.full([self.width + 2, self.height + 2], self.__defaultValue, dtype=np.int32)
            for sparseData in cellsData["data"]:
                x = int(sparseData[0])
                y = int(sparseData[1])
                value = int(sparseData[2])
                self.__cells[x, y] = value

        cellsValues = cellsData["values"]
        for serialKey, serialValue in cellsValues.items():
            serialId = int(serialKey)
            value = CellValue.fromName(serialValue)
            self.__cells[self.__cells == serialId] = value

        unitsData = {}
        for unitId, unitData in data["units"]["data"].items():
            unit = Unit()
            unit.takeData(unitData)
            unitsData[int(unitId)] = unit

        self.__units = {}
        for unitData in data["units"]["cells"]:
            if type(unitData) == dict:
                x = int(unitData["x"])
                y = int(unitData["y"])
                unitId = int(unitData["id"])
            elif type(unitData) == list:
                x = int(unitData[0])
                y = int(unitData[1])
                unitId = int(unitData[2])
            else:
                raise ValueError(f"Invalid unit data type {type(unitData)}")
            self.__units[(x, y)] = unitsData[unitId]
