from typing import Tuple

from core.constants import UnitProperty
from core.constants.UnitClass import unitClassId2Name, UnitClass
from core.logic import Logic
from core.state import Unit, IGameStateListener, GameState
from .FrameComponent import FrameComponent
from ..control.Icon import Icon
from ..control.Label import Label
from ...theme.Theme import Theme


class UnitFrame(FrameComponent, IGameStateListener):

    def __init__(self, theme: Theme, logic: Logic, unit: Unit):
        super().__init__(theme)
        self.__logic = logic
        self.__unit = unit
        self.__unitsLayer = logic.world.units

        # Icon
        if unit is not None:
            unitClass = unit.unitClass
            playerId = unit.playerId
        else:
            unitClass = UnitClass.NONE
            playerId = 0
        unitsTileset = theme.getTileset("units")
        classTile = unitsTileset.getTile(unitClass, playerId)
        self.__icon = Icon(theme, classTile)
        self.__icon.moveRelativeTo("topLeft", self, "topLeft")
        self.addComponent(self.__icon)
        
        title, pointsDesc, propertiesDesc = self.__createDescriptions()

        # Title
        self.__titleLabel = Label(theme, title)
        self.__titleLabel.moveRelativeTo("topLeft", self.__icon, "topRight")
        self.addComponent(self.__titleLabel)

        # Action points
        self.__pointsLabel = Label(theme, pointsDesc)
        self.__pointsLabel.moveRelativeTo("topLeft", self.__titleLabel, "bottomLeft")
        self.addComponent(self.__pointsLabel)

        # Description
        self.__propertiesLabel = Label(theme, propertiesDesc)
        self.__propertiesLabel.moveRelativeTo("topLeft", self.__pointsLabel, "topRight")
        self.addComponent(self.__propertiesLabel)

        self.pack()

        # Listeners
        logic.state.registerListener(self)

    def dispose(self):
        self.__logic.state.removeListener(self)

    def __createDescriptions(self) -> Tuple[str, str, str]:
        unit = self.__unit
        if unit is None:
            return "No unit", "", ""

        title = f'{unitClassId2Name[unit.unitClass]}<flag player="{unit.playerId}">'

        pointsDesc = ""
        # Life points
        if unit.hasProperty(UnitProperty.LIFE_POINTS) \
                and unit.hasProperty(UnitProperty.MAX_LIFE_POINTS):
            value = unit.getProperty(UnitProperty.LIFE_POINTS)
            maxValue = unit.getProperty(UnitProperty.MAX_LIFE_POINTS)
            pointsDesc += f"{value}/{maxValue}<life><br>"
        # Move points
        if unit.hasProperty(UnitProperty.MOVE_POINTS) \
                and unit.hasProperty(UnitProperty.MAX_MOVE_POINTS):
            value = unit.getProperty(UnitProperty.MOVE_POINTS)
            maxValue = unit.getProperty(UnitProperty.MAX_MOVE_POINTS)
            pointsDesc += f"{value}/{maxValue}<move><br>"
        # Attack points
        if unit.hasProperty(UnitProperty.ACTION_POINTS) \
                and unit.hasProperty(UnitProperty.MAX_ACTION_POINTS):
            value = unit.getProperty(UnitProperty.ACTION_POINTS)
            maxValue = unit.getProperty(UnitProperty.MAX_ACTION_POINTS)
            pointsDesc += f"{value}/{maxValue}<attack><br>"
            
        # Attack
        propertiesDesc = ""
        if unit.hasProperty(UnitProperty.MELEE_ATTACK):
            value = unit.getProperty(UnitProperty.MELEE_ATTACK)
            propertiesDesc += f"{value}<melee_attack>"
        if unit.hasProperty(UnitProperty.BOW_ATTACK):
            value = unit.getProperty(UnitProperty.BOW_ATTACK)
            propertiesDesc += f"{value}<range_attack>"
        if unit.hasProperty(UnitProperty.MOUNT_ATTACK):
            value = unit.getProperty(UnitProperty.MOUNT_ATTACK)
            propertiesDesc += f"{value}<mount_attack>"
        if unit.hasProperty(UnitProperty.SIEGE_ATTACK):
            value = unit.getProperty(UnitProperty.SIEGE_ATTACK)
            propertiesDesc += f"{value}<siege_attack>"
        propertiesDesc += "<br>"

        # Defense
        if unit.hasProperty(UnitProperty.MELEE_DEFENSE):
            value = unit.getProperty(UnitProperty.MELEE_DEFENSE)
            propertiesDesc += f"{value}<melee_defense>"
        if unit.hasProperty(UnitProperty.BOW_DEFENSE):
            value = unit.getProperty(UnitProperty.BOW_DEFENSE)
            propertiesDesc += f"{value}<range_defense>"
        propertiesDesc += "<br>"

        # Defense
        if unit.hasProperty(UnitProperty.MOUNT_DEFENSE):
            value = unit.getProperty(UnitProperty.MOUNT_DEFENSE)
            propertiesDesc += f"{value}<mount_defense>"
        if unit.hasProperty(UnitProperty.SIEGE_DEFENSE):
            value = unit.getProperty(UnitProperty.SIEGE_DEFENSE)
            propertiesDesc += f"{value}<siege_defense>"
        propertiesDesc += "<br>"

        return title, pointsDesc, propertiesDesc

    # GameState listener

    def unitChanged(self, state: GameState, cell: Tuple[int, int], unit: Unit):
        if unit != self.__unit:
            return
        title, actionDesc, propDesc = self.__createDescriptions()
        self.__titleLabel.setMessage(title)
        self.__pointsLabel.setMessage(actionDesc)
        self.__propertiesLabel.setMessage(propDesc)
